#include "streamline_c.h"

char streamline_debug;

void streamline(int nx, int ny,
		float *u, float *v, float *xx, float *yy,
		float dl, int idensity, struct StreamOpt opt)
{
  float nstart,nstart2;
  float sx,sy;
  int flag,k;
  float *lines_x,*lines_y;
  int *lines_id;
  int lines_pos,id;

  /*allocate work space*/
  lines_x = malloc(sizeof(float)*opt.lines_max);
  if (lines_x == NULL){
    if (streamline_debug) printf("STREAMLINE: cannot allocate work space\n");
    return;
  }
  lines_y = malloc(sizeof(float)*opt.lines_max);
  if (lines_y == NULL){
    if (streamline_debug) printf("STREAMLINE: cannot allocate work space\n");
    return;
  }
  lines_id = malloc(sizeof(int)*opt.lines_max);
  if (lines_id == NULL){
    if (streamline_debug) printf("STREAMLINE: cannot allocate work space\n");
    return;
  }
  id = 0;

  if (nx>ny)
    nstart = ((float)nx / (float)idensity);
  else
    nstart = ((float)ny / (float)idensity);
  nstart2 = nstart*nstart;

  /*temporary setting*/
  //opt.touch_thresh = nstart / 3.0;

  lines_pos = 0;

  for (sx=(nstart/5.0);sx<nx;sx+=(nstart/5.0)){
    for (sy=(nstart/5.0);sy<ny;sy+=(nstart/5.0)){
      flag = 1;
      for (k=0;k<(lines_pos);k++){
	if( (lines_x[k]-sx)*(lines_x[k]-sx)
	    +(lines_y[k]-sy)*(lines_y[k]-sy) <= nstart2 ){
	  flag = 0;
	  break;
	}
      }
      if (flag){
	streamline_core(nx, ny, u, v, xx, yy, dl, sx, sy,
			lines_x, lines_y, lines_id,
			&lines_pos, id++, opt);
      }
    }
  }

  /*free work space*/
  free(lines_x);
  free(lines_y);
}

void streamline_core(int nx, int ny,
		     float *u, float *v, float *xx, float *yy,
		     float dl, float sx, float sy,
		     float *lines_x, float *lines_y, int *lines_id,
		     int *lines_pos, int id, struct StreamOpt opt)
{
#define LINEMAX 10000
  float t,dt;
  float x,y;
  float u0,v0;
  float dx,dy;
  float xx0,xx1,yy0,yy1;
  float thisline_x[2][LINEMAX],thisline_y[2][LINEMAX];
  float thisline_xx[LINEMAX],thisline_yy[LINEMAX];
  int thisline_pos[2];
  integer thisline_pos2;
  int flag,loopflag,loopend;
  char crossflag;
  int i,j;
  float su,sv;
  int index,ltype;
  float arrowsize;
  integer false=0,len_cp=5;

  interp_2d_vect(nx,ny,u,v,sx,sy,&su,&sv);
  if (streamline_debug) printf("STREAMLINE: new line\n");
  thisline_pos[0] = 0;
  thisline_pos[1] = 0;
  for (j=0;j<2;j++){
    thisline_pos2 = 0;
    if(j==0){
      dt = -0.1;
      t = opt.tarrow;
      if (streamline_debug) printf("STREAMLINE: backward\n");
    }
    else{
      dt = 0.1;
      t = 0.0;
      if (streamline_debug) printf("STREAMLINE: forward\n");
    }

    x = sx;
    y = sy;
    if(interp_2d_vect(nx,ny,xx,yy,x,y,&xx0,&yy0)){
      //if (streamline_debug) printf("STREAMLINE: out of range (in interp_2d_vect, 0):x=%f,y=%f\n",x,y);
      if (streamline_debug) printf("STREAMLINE: initial position is out of range\n");
      break;
    }
    if(j==0){
      thisline_x[0][LINEMAX - thisline_pos[0] - 1] = x;
      thisline_y[0][LINEMAX - thisline_pos[0] - 1] = y;
    }
    else{
      thisline_x[1][thisline_pos[1]] = x;
      thisline_y[1][thisline_pos[1]] = y;
    }
    thisline_pos[j]++;

    thisline_xx[thisline_pos2] = xx0;
    thisline_yy[thisline_pos2] = yy0;
    thisline_pos2++;

    flag = 1;
    loopflag = 0;
    while(flag){
      if(interp_2d_vect(nx,ny,u,v,x,y,&u0,&v0)){
	if (streamline_debug) printf("STREAMLINE: out of range (in interp_2d_vect)\n");
	break;
      }
      
      /*exit if array thisline_? is full*/
      if (thisline_pos[j] == LINEMAX){
	uulinz_(&thisline_pos2,&thisline_xx,&thisline_yy,
		&(opt.type),&(opt.index));
	if (streamline_debug) printf("STREAMLINE: line length exceeds maximum\n");
	/*
	for(i=(*lines_pos);i<(*lines_pos)+thisline_pos;i++){
	  if (i>=opt.lines_max){
	    printf("STREAMLINE: work space is not enough\n");
	    return;
	  }
	  lines_x[i]=thisline_x[i-(*lines_pos)];
	  lines_y[i]=thisline_y[i-(*lines_pos)];      
	}
	*lines_pos += thisline_pos;
	*/
	if((*lines_pos) + thisline_pos[0] + thisline_pos[1] > opt.lines_max){
	  if (streamline_debug) printf("STREAMLINE: work space is not enough\n");
	  return;
	}
	if(thisline_pos[0]+thisline_pos[1]>2){
	  memcpy(lines_x + (*lines_pos), 
		 (float*)thisline_x + (LINEMAX - thisline_pos[0]), 
		 ((thisline_pos[0] + thisline_pos[1]) * sizeof(float)));
	  memcpy(lines_y + (*lines_pos), 
		 (float*)thisline_y + (LINEMAX - thisline_pos[0]), 
		 ((thisline_pos[0] + thisline_pos[1]) * sizeof(float)));
	  if (streamline_debug) printf("STREAMLINE: id=%d\n",id);
	  for(i=(*lines_pos);i<((*lines_pos)+thisline_pos[0]+thisline_pos[1]);i++)
	    lines_id[i] = id;
	  (*lines_pos) += (thisline_pos[0]+thisline_pos[1]);
	  return;
	}
      }

      /*check for stagnation*/
      if (u0*u0 + v0*v0 < (opt.slow_thresh)*(opt.slow_thresh)){
	if (streamline_debug) printf("STREAMLINE: stagnation detected\n");
	break;
      }

      /*integration*/
      if(rungekutta_2d(nx,ny,u,v,x,y,dt/dl,&dx,&dy)){
	if (streamline_debug) printf("STREAMLINE: out of range (in rungekutta, 1)\n");
	break;
      }
      
      /*modify dt*/
      if ((fabs(dx) < opt.dt_thresh0) && (fabs(dy) < opt.dt_thresh0)){
	if (fabs(dx) > fabs(dy)){
	  dt *= (opt.dt_thresh0/fabs(dx) * 1.1);
	}
	else{
	  dt *= (opt.dt_thresh0/fabs(dy) * 1.1);
	}
	if(rungekutta_2d(nx,ny,u,v,x,y,dt/dl,&dx,&dy)){
	  if (streamline_debug) printf("STREAMLINE: out of range (in rungekutta, 2)\n");
	  break;
	}
      }
      while (fabs(dx) > opt.dt_thresh1 && fabs(dy) > opt.dt_thresh1){
	if (fabs(dx) > fabs(dy)){
	  dt *= (opt.dt_thresh1/fabs(dx) * 0.9);
	}
	else{
	  dt *= (opt.dt_thresh1/fabs(dy) * 0.9);
	}
	if(rungekutta_2d(nx,ny,u,v,x,y,dt/dl,&dx,&dy)){
	  flag = 0;
	  if (streamline_debug) printf("STREAMLINE: out of range (in rungekutta, 3)\n");
	  break;
	}
      }
      if(!flag) break;
      
      /*break if point is out of the region*/
      if (x + dx < 0.0 || x + dx > nx-1 || y + dy < 0.0 || y + dy > ny-1){
	if (streamline_debug) printf("STREAMLINE: out of range\n");
	break;
      }

      /*
      if(interp_2d_vect(nx,ny,xx,yy,x,y,&xx0,&yy0)){
	if (streamline_debug) printf("STREAMLINE: out of range (in interp_2d_vect)\n");
	break;
      }
      */
      if(interp_2d_vect(nx,ny,xx,yy,x+dx,y+dy,&xx1,&yy1)){
	if (streamline_debug) printf("STREAMLINE: out of range (in interp_2d_vect)\n");
	break;
      }

      
      /*draw arrows*/
      t += fabs(dt);
      if (t>=opt.tarrow){
	sglset_("LPROP",&false,(integer)strlen("LPROP"));
	//if ((fabs(u0)>1) && (fabs(v0)>1))
	//arrowsize = 0.01;
	//else
	//arrowsize = 0.005;
	sgrset_("CONST",&(opt.arrowsize),(integer)strlen("CONST"));
	sgrset_("CONSTM",&(opt.arrowsize_map),(integer)strlen("CONSTM"));
	if (dt>=0)
	  sglazu_(&xx0,&yy0,&xx1,&yy1,&(opt.type),&(opt.index));
	else
	  sglazu_(&xx1,&yy1,&xx0,&yy0,&(opt.type),&(opt.index));
	t = t - opt.tarrow;
      }
      /*
      else
	sglnzu_(&xx0,&yy0,&xx1,&yy1,&(opt.index));
      */

      /*update current position*/
      x += dx;
      y += dy;
      /*update trace*/
      if(j==0){
	thisline_x[0][LINEMAX - thisline_pos[0] - 1] = x;
	thisline_y[0][LINEMAX - thisline_pos[0] - 1] = y;
      }
      else{
	thisline_x[1][thisline_pos[1]] = x;
	thisline_y[1][thisline_pos[1]] = y;
      }
      thisline_pos[j]++;
      /*update current position in U-coordinate*/
      xx0 = xx1;
      yy0 = yy1;
      /*update trace in U-coordinate*/
      thisline_xx[thisline_pos2] = xx1;
      thisline_yy[thisline_pos2] = yy1;
      thisline_pos2++;

      /*check for other lines*/
      crossflag = 0;
      for (i=0;i<(*lines_pos);i++){
	/*contact*/
	if ((lines_x[i]-x)*(lines_x[i]-x) 
	    + (lines_y[i]-y)*(lines_y[i]-y) 
	    <= opt.touch_thresh*opt.touch_thresh){
	  flag = 0;
	  if (streamline_debug) printf("STREAMLINE: contact detected\n");
	  break;
	}
	/*crossline*/
	if((i+1)<(*lines_pos)){
	  if(lines_id[i] == lines_id[i+1]){
	    if((((x - lines_x[i]) * dy - (y - lines_y[i]) * dx)
		* ((x - lines_x[i+1]) * dy - (y - lines_y[i+1]) * dx) < 0) &&
	       (((x - lines_x[i])*(lines_y[i+1] - lines_y[i])
		 - (y - lines_y[i])*(lines_x[i+1] - lines_x[i]))
		* ((x + dx - lines_x[i])*(lines_y[i+1] - lines_y[i])
		   - (y - lines_y[i])*(lines_x[i+1] - lines_x[i])) < 0)){
	      crossflag = 1;
	      if (streamline_debug) 
		printf("STREAMLINE: crossline detected: target id=%d, (%03.1f,%03.1f)\n",lines_id[i-1],lines_x[i-1],lines_y[i-1]);
		printf("STREAMLINE:                     target id=%d, (%03.1f,%03.1f)\n",lines_id[i],lines_x[i],lines_y[i]);
		printf("STREAMLINE:                     target id=%d, (%03.1f,%03.1f)\n",lines_id[i+1],lines_x[i+1],lines_y[i+1]);
		printf("STREAMLINE:                     target id=%d, (%03.1f,%03.1f)\n",lines_id[i+2],lines_x[i+2],lines_y[i+2]);
		printf("STREAMLINE:                     current data  (%3.1f,%3.1f)-(%3.1f,%3.1f)\n",x,y,x+dx,y+dy);
		//printf("STREAMLINE:                     (%3.1f,%3.1f)-(%3.1f,%3.1f)\n",
		//       lines_x[i],lines_y[i],lines_x[i+1],lines_y[i+1]);
	      //printf("%f,%f,%f,%f\n",x,x-dx,lines_x[i],lines_x[i+1]);
	      //printf("%f,%f,%f,%f\n",y,y-dy,lines_y[i],lines_y[i+1]);
	      break;
	    }
	  }
	}
      }
      if(crossflag) break;
      
      /*check for loops of this line*/
      if (loopflag && (thisline_pos[0]+thisline_pos[1]>=2)){
	if (thisline_pos[0]+thisline_pos[1] < 10){
	  loopend = (thisline_pos[0]+thisline_pos[1]) / 2.0;
	}
	else{
	  loopend = (thisline_pos[0]+thisline_pos[1]) - 10.0;
	}
	if(j==0){
	  for (i=thisline_pos[0]-loopend+1 ;i<=thisline_pos[0];i++){
	    if ((thisline_x[0][LINEMAX-i]-x)*(thisline_x[0][LINEMAX-i]-x) 
		+ (thisline_y[0][LINEMAX-i]-y)*(thisline_y[0][LINEMAX-i]-y) <= opt.loop_thresh){
	      flag = 0;
	      if (streamline_debug) printf("STREAMLINE: loop detected\n");
	      break;
	    }
	  }
	}
	else{
	  for (i=thisline_pos[1]-loopend+1 ;i<=thisline_pos[1];i++){
	    if (((*((float*)thisline_x+(LINEMAX+i))) - x)*((*((float*)thisline_x+(LINEMAX+i))) - x) 
		+ ((*((float*)thisline_y+(LINEMAX+i))) - y)*((*((float*)thisline_y+(LINEMAX+i))) - y) <= opt.loop_thresh){
	      flag = 0;
	      if (streamline_debug) printf("STREAMLINE: loop detected\n");
	      break;
	    }
	  }
	}
      }

      /*check for loop trigger*/
      //printf("%f,%f,%f,%f\n",su,sv,u0,v0);
      if (u0*su+v0*sv < 0) loopflag = 1;
      //if (u0*su+v0*sv < 0) printf("loop\n");

      /*escape from dead lock*/
      if (thisline_pos[j] >= 2)
	if ((thisline_x[j][thisline_pos[j]-2]-x)*(thisline_x[j][thisline_pos[j]-2]-x)
	    + (thisline_y[j][thisline_pos[j]-2]-y)*(thisline_y[j][thisline_pos[j]-2]-y) 
	    <= 10e-10){ 
	  if (streamline_debug) printf("STREAMLINE: dead lock detected\n");
	  flag = 0;
	}
    } 
    /*draw streamline*/
    if(thisline_pos2>2)
      uulinz_(&thisline_pos2,&thisline_xx,&thisline_yy,
	      &(opt.type),&(opt.index));
    /*
    if(thisline_pos2>2)
      uumrk_(&thisline_pos2,&thisline_xx,&thisline_yy);
    */
  }
  if (streamline_debug) printf("STREAMLINE: thisline_pos=%d,%d\n",thisline_pos[0],thisline_pos[1]);
  /*export line trace*/
  if((*lines_pos) + thisline_pos[0] + thisline_pos[1] > opt.lines_max){
    if (streamline_debug) printf("STREAMLINE: work space is not enough\n");
    return;
  }
  if(thisline_pos[0]+thisline_pos[1]>2){
    memcpy(lines_x + (*lines_pos), 
	   (float*)thisline_x + (LINEMAX - thisline_pos[0]), 
	   ((thisline_pos[0] + thisline_pos[1]) * sizeof(float)));
    memcpy(lines_y + (*lines_pos), 
	   (float*)thisline_y + (LINEMAX - thisline_pos[0]), 
	   ((thisline_pos[0] + thisline_pos[1]) * sizeof(float)));
    if (streamline_debug) printf("STREAMLINE: id=%d\n",id);
    for(i=(*lines_pos);i<((*lines_pos)+thisline_pos[0]+thisline_pos[1]);i++)
      lines_id[i] = id;
    (*lines_pos) += (thisline_pos[0]+thisline_pos[1]);
  }
  return;
}

int rungekutta_2d(int nx, int ny,
		  float *u, float *v,
		  float x0, float y0, 
		  float dt, float *dx, float *dy)
{
  float u0,v0,kx,ky;
  
  if(interp_2d_vect(nx,ny,u,v,x0,y0,&u0,&v0))
    return 1;
  kx = u0 * dt;
  ky = v0 * dt;
  *dx = kx / 6.0;
  *dy = ky / 6.0;

  if(interp_2d_vect(nx,ny,u,v,x0+kx/2.0,y0+ky/2.0,&u0,&v0))
    return 1;
  kx = u0 * dt;
  ky = v0 * dt;
  *dx += kx / 3.0;
  *dy += ky / 3.0;

  if(interp_2d_vect(nx,ny,u,v,x0+kx/2.0,y0+ky/2.0,&u0,&v0))
    return 1;
  kx = u0 * dt;
  ky = v0 * dt;
  *dx += kx / 3.0;
  *dy += ky / 3.0;

  if(interp_2d_vect(nx,ny,u,v,x0+kx,y0+ky,&u0,&v0))
    return 1;
  kx = u0 * dt;
  ky = v0 * dt;
  *dx += kx / 6.0;
  *dy += ky / 6.0;

  if(x0+(*dx)<0 || x0+(*dx)>nx-1 || y0+(*dy)<0 || y0+(*dy)>ny-1) return 1;

  return 0;
}

int interp_2d_vect(int nx, int ny,
		   float *u, float *v,
		   float x, float y,
		   float *u0, float *v0)
{
  int x1,x2,y1,y2;
  float dx1,dx2,dy1,dy2;

  if(x<0 || x>nx-1 || y<0 || y>ny-1) return 1;
  /*
  if (x<0) return 1;
  if (x>nx-1) return 2;
  if (y<0) return 3;
  if (y>ny-1) return 4;

  if (x<0) x=0;
  if (x>nx-1) x=nx-1;
  if (y<0) y=0;
  if (y>ny-1) y=ny-1;
  */
  x1 = floor(x); 
  x2 = ceil(x);
  y1 = floor(y); 
  y2 = ceil(y);
  dx1 = x - x1; 
  dx2 = 1 - dx1; //x2 - x;
  dy1 = y - y1; 
  dy2 = 1 - dy1; //y2 - y;
  if (x1 == x2){
    if (y1 == y2){
      *u0 = u[x1+y1*nx];
      *v0 = v[x1+y1*nx];
    }
    else{
      *u0 = u[x1+y2*nx]*dy1 + u[x1+y1*nx]*dy2;
      *v0 = v[x1+y2*nx]*dy1 + v[x1+y1*nx]*dy2;
    }
  }
  else{
    if (y1 == y2){
      *u0 = u[x2+y1*nx]*dx1 + u[x1+y1*nx]*dx2;
      *v0 = v[x2+y1*nx]*dx1 + v[x1+y1*nx]*dx2;
    }
    else{
      *u0 = (u[x2+y2*nx]*dx1 + u[x1+y2*nx]*dx2)*dy1;
      *u0 +=(u[x2+y1*nx]*dx1 + u[x1+y1*nx]*dx2)*dy2;
      *v0 = (v[x2+y2*nx]*dx1 + v[x1+y2*nx]*dx2)*dy1;
      *v0 +=(v[x2+y1*nx]*dx1 + v[x1+y1*nx]*dx2)*dy2;
    }
  }
  return 0;
}
