#include "ruby.h"
#include "narray.h"
#include "streamline_c.h"

static VALUE streamline_ruby2c(self,u,v,xx,yy,dl,idensity,type,index)
     VALUE self,u,v,xx,yy,dl,idensity,type,index;
{
  VALUE u1,v1,xx1,yy1,dl1,id1;
  struct NARRAY *nu,*nv,*nxx,*nyy,*ndl,*nid;
  float *pu,*pv,*pxx,*pyy;
  float fdl;
  int iid,itype,iindex;
  struct StreamOpt opt;
  int num_x,num_y;

  //printf("start streamline_ruby2c\n");

  //printf("u \n");
  if NA_IsNArray(u){
    u1 = na_cast_object(u, NA_SFLOAT);
    GetNArray(u1,nu);
    pu = (float*)nu->ptr;
    //printf("get num_x\n");
    num_x = nu->shape[0];
    num_y = nu->shape[1];
    //printf("num_x=%d\n",num_x);
  }
  else{
    rb_raise(rb_eTypeError,"1st arg should be a NArray");
    return Qfalse;
  }
  
  if NA_IsNArray(v){
    //printf("v \n");
    v1 = na_cast_object(v, NA_SFLOAT);
    GetNArray(v1,nv);
    pv = (float*)nv->ptr;
  }
  else{
    rb_raise(rb_eTypeError,"2nd arg should be a NArray");
    return Qfalse;
  }
  if(nv->total != nu->total)
    rb_raise(rb_eArgError,
	     "array size of 2nd arg is different from that of 1st arg");

  if NA_IsNArray(xx){
    //printf("xx \n");
    xx1 = na_cast_object(xx, NA_SFLOAT);
    GetNArray(xx1,nxx);
    pxx = (float*)nxx->ptr;
  }
  else{
    rb_raise(rb_eTypeError,"3rd arg should be a NArray");
    return Qfalse;
  }
  if(nv->total != nu->total)
    rb_raise(rb_eArgError,
	     "array size of 3rd arg is different from that of 1st arg");

  if NA_IsNArray(yy){
    //printf("yy \n");
    yy1 = na_cast_object(yy, NA_SFLOAT);
    GetNArray(yy1,nyy);
    pyy = (float*)nyy->ptr;
  }
  else{
    rb_raise(rb_eTypeError,"4th arg should be a NArray");
    return Qfalse;
  }
  if(nv->total != nu->total)
    rb_raise(rb_eArgError,
	     "array size of 4th arg is different from that of 1st arg");

  if NA_IsNArray(dl){
    //printf("dl \n");
    dl1 = na_cast_object(dl, NA_SFLOAT);
    GetNArray(dl1,ndl);
    fdl = *(float*)ndl->ptr;
    //printf("dl=%f\n",*fdl);
  }
  else{
    switch(TYPE(dl)){
    case T_FLOAT:
      //fdl = (double)(RFLOAT(dl)->value);
      fdl = NUM2DBL(dl);
      break;
    case T_FIXNUM:
      fdl = FIX2INT(dl);
      break;
    defalut1:
      rb_raise(rb_eTypeError,"dl should be a NArray or a Float or a Fixnum");
      return Qfalse;
      break;
    }
  }

  if NA_IsNArray(idensity){
    id1 = na_cast_object(idensity, NA_LINT);
    GetNArray(id1,nid);
    iid = (int)(nid->ptr[0]);
  }
  else{
    switch(TYPE(idensity)){
    case T_FLOAT:
      //iid = (double)(RFLOAT(idensity)->value);
      iid = NUM2DBL(idensity);
      break;
    case T_FIXNUM:
      iid = FIX2INT(idensity);
      break;
    defalut2:
      rb_raise(rb_eTypeError,"idensity should be a NArray or a Float or a Fixnum");
      return Qfalse;
      break;
    }
  }

  switch(TYPE(type)){
  case T_FLOAT:
    //itype = (double)(RFLOAT(type)->value);
    itype = NUM2DBL(type);
    break;
  case T_FIXNUM:
    itype = FIX2INT(type);
    break;
  defalut3:
    rb_raise(rb_eTypeError,"type should be a Float or a Fixnum");
    return Qfalse;
    break;
  }

  switch(TYPE(index)){
  case T_FLOAT:
    //iindex = (double)(RFLOAT(index)->value);
    iindex = NUM2DBL(index);
    break;
  case T_FIXNUM:
    iindex = FIX2INT(index);
    break;
  defalut4:
    rb_raise(rb_eTypeError,"type should be a Float or a Fixnum");
    return Qfalse;
    break;
  }

  /*set options*/
  opt.dt_thresh0 = 0.1;
  opt.dt_thresh1 = 0.3;
  opt.loop_thresh = 1.0;//0.05;
  if(num_x > num_y)
    opt.touch_thresh = ((float)num_x) / ((float)iid) * 0.5;
  else
    opt.touch_thresh = ((float)num_y) / ((float)iid) * 0.5;
  if(opt.touch_thresh < 0.1)
    opt.touch_thresh = 0.1;
  //printf("opt.touch_thresh=%f\n",opt.touch_thresh);
  opt.slow_thresh = 0.1;
  opt.tarrow = (60.0*60.0*48.0);
  opt.arrowsize = 0.01;
  opt.arrowsize_map = 0.5;
  opt.lines_max = 1000000;
  opt.type = itype;
  opt.index = iindex;

  /*debug flag*/
  if(ruby_debug) streamline_debug = 1;

  /*call main routine*/
  streamline(num_x,num_y,pu,pv,pxx,pyy,fdl,iid,opt);

  return Qnil;
}

void Init_dclext_streamline()
{
  VALUE mNumRu,mDCLEXT;
  mNumRu = rb_define_module("NumRu");
  mDCLEXT = rb_define_module_under(mNumRu, "DCLExt");
  rb_define_module_function(mDCLEXT, "streamline_c", streamline_ruby2c, 8);
}
